unit Editor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, Menus,utils, mwCustomEdit, mwHighlighter,
  typinfo, DFMUtils, ImgList, dmDfmSyn, mwPasSyn, mwCompletionProposal;

type

  TWrAction= (wrInsert, wrDelete, wrRename);
  TUnitState=(usNone, usNew,usOpen,usSave,usChange);

  TParserUnit = class(TObject)
  private
     fUnitName:string;
     fUnitState : TUnitState;
     fUnitSource:TStringList;
     procedure SetUnitName(FileName:string);
  protected
     function  GetStrPos(const Str:string):Integer;
  public
     constructor Create(const FileName:string);
     destructor Destroy; override;
     procedure SetUnitSource(List: TStrings);
     property UnitName : string read fUnitName write SetUnitName;
     property UnitState : TUnitState read fUnitState write fUnitState;
     property UnitSource : TStringList read fUnitSource write fUnitSource;
  end;

  TEditorForm = class(TForm)
    StatusBar: TStatusBar;
    PageControl1: TPageControl;
    OpenDialog1: TOpenDialog;
    MsgViewPopupMenu: TPopupMenu;
    mvViewSourceItem: TMenuItem;
    mvAltViewSourceItem: TMenuItem;
    mvEditSourceItem: TMenuItem;
    mvAltEditSourceItem: TMenuItem;
    N3: TMenuItem;
    ClearCompilerMessagesItem: TMenuItem;
    ClearSearchResultsItem: TMenuItem;
    EditorLocalMenu: TPopupMenu;
    ecClosePage: TMenuItem;
    ecOpenFileAtCursor: TMenuItem;
    ecNewEditWindow: TMenuItem;
    ecBrowseSymbolAtCursor: TMenuItem;
    ecTopicSearch: TMenuItem;
    ecAddtoInterface: TMenuItem;
    ecToggleBreakpoint: TMenuItem;
    ecGotoCursor: TMenuItem;
    ecEvaluateModify: TMenuItem;
    ecAddWatchAtCursor: TMenuItem;
    MenuItem1: TMenuItem;
    ecViewAsFormItem: TMenuItem;
    ecReadOnlyItem: TMenuItem;
    ecMessageView: TMenuItem;
    MenuItem2: TMenuItem;
    ecPropertiesItem: TMenuItem;
    SmallGutterImages: TImageList;
    LargeGutterImages: TImageList;
    PasSyn: TmwPasSyn;
    N1: TMenuItem;
    ToggleBookmakrs1: TMenuItem;
    GotoBookamrks1: TMenuItem;
    Bookamrk01: TMenuItem;
    Bookmark11: TMenuItem;
    Bookmark21: TMenuItem;
    Bookmark31: TMenuItem;
    Bookmark41: TMenuItem;
    Bookmark51: TMenuItem;
    Bookmark61: TMenuItem;
    Bookmark71: TMenuItem;
    Bookmark81: TMenuItem;
    Bookmark91: TMenuItem;
    Bookmark01: TMenuItem;
    Bookmark12: TMenuItem;
    Bookmark22: TMenuItem;
    Bookmark32: TMenuItem;
    Bookmark42: TMenuItem;
    Bookmark52: TMenuItem;
    Bookmark62: TMenuItem;
    Bookmark72: TMenuItem;
    Bookmark82: TMenuItem;
    Bookmark92: TMenuItem;
    DfmSyn: TdmDfmSyn;
    MwCompletionProposal1: TMwCompletionProposal;
    mwAutoComplete1: TmwAutoComplete;
    ecViewExplorerItem: TMenuItem;
    ecViewCodeTemplate: TMenuItem;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure PageControl1Change(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ecClosePageClick(Sender: TObject);
    procedure ecOpenFileAtCursorClick(Sender: TObject);
    procedure ecReadOnlyItemClick(Sender: TObject);
    procedure Bookamrk01Click(Sender: TObject);
    procedure Bookmark01Click(Sender: TObject);
    procedure ecTopicSearchClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ecViewAsFormItemClick(Sender: TObject);
    procedure MwCompletionProposal1Execute(Sender: TObject);
    procedure PasSynToken(Sender: TObject; TokenKind: Integer;
      TokenText: String; LineNo: Integer);
    procedure ecViewExplorerItemClick(Sender: TObject);
    procedure ecViewCodeTemplateClick(Sender: TObject);
    procedure MwCompletionProposal1CodeCompletion(var Value: String;
      var X: Integer);
  private
    { Private declarations }
    ParserUnit: TParserUnit;
    function GetStringPos(const S:string; WholeWord: Boolean):Integer;
    procedure CaretInfo;
    function  IsIncludeUnit(UnitName:string): Boolean;
    procedure ReplaceSelText(Replace:string);
    function  GetEventHandlerPosition(const EventName:string):Integer;
  protected
    procedure OnGetMinMaxInfo(var msg : TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure CodeTemplateValidate(Sender: TObject);
  public
    { Public declarations }
     procedure WriteComponentCode(Component:string;ComponentClass:
               TComponentClass;Action:TWrAction);
     procedure WriteEventCode(Event:string);
     procedure GotoLine(PasEdit:TmwCustomEdit;Line:Integer);
     function  GetLinkEditor:TmwCustomEdit;
     function  SetFocusEventHandler(const EventName:string):Boolean;
     procedure RenameUnit(UnitName:string);
     procedure NewSheet(const FileName:string);
     function  GetCursorWord:string;
     function  GetLinkWithFile(const FileName:string): TmwCustomEdit;
     procedure CheckUseUnit(Component: TComponent);
    procedure  AddUseUnit(UnitName:string);
    procedure StatusChangeHandler(Sender: TObject; Changes: TmwStatusChanges);
  end;

 TParseState = (psIdle, psInClassDef, psProcSeen, psWaitForSymbol);

var
  EditorForm: TEditorForm;
  ParseState: TParseState;
  ProcedureName: string;
  CompletionProposalForm: TCompletionProposalForm;

implementation

uses MainForm,Uconst, utype, Proxy, ObjectInspec;

var
   Insert:boolean;

{$R *.DFM}

{TParserUnit}

constructor TParserUnit.Create(const FileName:string);
begin
   fUnitState:= usNone;
   FUnitSource:= TStringList.Create;
end;

destructor TParserUnit.Destroy;
begin
   FUnitSource.Free;
   inherited;
end;

function TParserUnit.GetStrPos(const Str:string):Integer;
// unit ҽ Ư ڿ ִ  ˷ش.
var
  I:Integer;
begin
   Result:=-1;
   for I:=0 to UnitSource.Count-1 do
      if Pos(Str,UnitSource[I])>0 then
      begin
         Result:=I;
         Exit;
      end;
end;

procedure TParserUnit.SetUnitName(FileName:string);
begin
   if FileName = '' then Exit;
   if fUnitName <> FileName then
   begin
      fUnitName:= FileName;
      if FileExists(FileName) then
         fUnitSource.LoadFromFile(FileName);
   end;
end;

procedure TParserUnit.SetUnitSource(List: TStrings);
begin
   if Assigned(List) then fUnitSource.Assign(List);
end;

{TEditorForm}
procedure TEditorForm.OnGetMinMaxInfo(var msg : TWMGetMinMaxInfo);
// Editor  ִȭ ,  ũ⸦    Ѵ.
begin
   inherited;
   msg.MinMaxInfo^.ptMaxPosition.y := Top;
   msg.MinMaxInfo^.ptMaxTrackSize.x := Screen.Width;
   msg.MinMaxInfo^.ptMaxTrackSize.y := Screen.Height -  Top;
end;

procedure TEditorForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
   I,SaveResp:integer;
   FileName:string;
begin
   for I := PageControl1.PageCount-1 downto 0 do
   begin
     FileName:=(PageControl1.Pages[I]).Caption;
     if (PageControl1.Pages[I].Controls[0] as TmwCustomEdit).Modified then
     begin
         SaveResp := MessageDlg(Format('Save Changes to %s?',
            [ExtractFilename(FileName)]), mtWarning, mbYesNoCancel, 0);
         case SaveResp of
            idYes: FMainForm.SaveFile(ProjectPath+Caption);
            idNo: ;
            idCancel: Abort;
         end;
     end;
   end;
end;

procedure TEditorForm.PageControl1Change(Sender: TObject);
begin
   // set caption to active filename
   if ExtractFileExt(PageControl1.ActivePage.Caption)= '' then
   // if active file is pascal file, then add extension '.pas'
      Caption:=PageControl1.ActivePage.Caption+'.pas'
   else Caption:=PageControl1.ActivePage.Caption;
   FMainForm.ActiveUnit:=
        FMainForm.OpenUnitList[PageControl1.ActivePage.Controls[0].Tag].FileName;
   ActivePasEditor:= TmwCustomEdit(PageControl1.ActivePage.Controls[0]);
end;

procedure TEditorForm.WriteComponentCode(Component:string;ComponentClass:
          TComponentClass;Action:TWrAction);
// Write/Modify code for component when created, deleted, modified
const
   ClassName= 'TForm';
var
   InsertPos: Integer;
   PasEdit: TmwCustomEdit;
begin
   PasEdit:= GetLinkEditor;
   if Action = wrInsert then
      InsertPos:= GetStringPos(ClassName, False)  // search position for inserting
   else
      InsertPos:= GetStringPos(Component, False);
   if InsertPos = 0 then Exit;
   case Action of
      wrInsert:  // insert code
      begin
          PasEdit.Lines.Insert(InsertPos+1,'    '+Component+':'+ComponentClass.ClassName+';');
      end;
      wrDelete:  // delete code
          PasEdit.Lines.Delete(InsertPos);
   end;
end;

function TEditorForm.GetStringPos(const S:string; WholeWord: Boolean):Integer;
// Return row position that include text S
var
   I: Integer;
   PasEdit: TmwCustomEdit;
begin
   Result:=0;
   PasEdit:= GetLinkEditor;
   with PasEdit do
   begin
      for I:=0 to Lines.Count-1 do
      begin
         if WholeWord then
         begin
            if CompareText(S, Trim(Lines[I])) = 0 then
            begin
              Result:= I;
              Exit;
            end;
         end
         else
         begin
           if Pos(S,Lines[I])>0 then
           begin
              Result:= I;
              Exit;
           end;
         end;
      end;
   end;
end;

function TEditorForm.GetEventHandlerPosition(const EventName:string):Integer;
// Return event handler's row position
var
   I: Integer;
   PasEdit: TmwCustomEdit;
begin
   Result:= 0;
   PasEdit:= GetLinkEditor;
   with PasEdit do
   begin
      for I:=Lines.Count - 1 downto 0 do
      begin
         if Pos(EventName, Lines[I])> 0 then
         begin
            Result:= I;
            Exit;
         end;
      end;
   end;
end;

// Write event handler code for Event
procedure TEditorForm.WriteEventCode(Event:string);
var
   PasEdit: TmwCustomEdit;
   InsertPos: Integer;
begin
   SetFocus;
   PasEdit:= GetLinkEditor;
   with PasEdit.Lines do
   begin
      // write event handler declare code
      InsertPos:= GetStringPos('private', True);
      if InsertPos <> 0 then
         Insert(InsertPos, Ident + Ident + 'procedure ' + Event);
      // write event handler definition code
      Insert(Count-2, '');
      Insert(Count-2, 'procedure '+'T'+ FMainForm.ActiveForm.Name+'.'+Event);
      Insert(Count-2, 'begin');
      Insert(Count-2, '');
      Insert(Count-2, 'end;');
   end;
   PasEdit.SetFocus;
   GotoLine(PasEdit, PasEdit.Lines.Count-3);
end;

procedure TEditorForm.GotoLine(PasEdit:TmwCustomEdit;Line:Integer);
begin
   // goto line number Line
   PasEdit.CaretX:= 0;
   PasEdit.CaretY:= Line;
end;

function TEditorForm.GetLinkEditor:TmwCustomEdit;
var
   I:Integer;
   FileName:string;
begin
   // Get active editor linked with active form
   Result:=TmwCustomEdit(PageControl1.ActivePage.Controls[0]);
   for I:=0 to PageControl1.PageCount-1 do
   begin
      FileName:= ExtractFileName(FMainForm.ActiveUnit);
      if TTabSheet(PageControl1.Pages[I]).Caption = FileName then
         Result:= TmwCustomEdit(PageControl1.Pages[I].Controls[0]);
   end;
end;

function TEditorForm.SetFocusEventHandler(const EventName:string):Boolean;
// Find event handler position at editor, then move caret to the position
var
   PasEdit: TmwCustomEdit;
   FindPos:Integer;
begin
   Result:= True;
   SetFocus;
   PasEdit:= GetLinkEditor;
   FindPos:= GetEventHandlerPosition(EventName);
   PasEdit.SetFocus;
   GotoLine(PasEdit, FindPos + 3);
   Result:= (FindPos <> 0);
end;

procedure TEditorForm.RenameUnit(UnitName:string);
// Rename active unit name to UnitName
var
  PasEdit:TmwCustomEdit;
  Name:string;
  Po: Integer;
begin
   PasEdit:= GetLinkWithFile(UnitName);
   if PasEdit = nil then Exit;
   Po:= ParserUnit.GetStrPos('unit');
   if Po = -1 then Exit;
   Name:= PasEdit.Lines[Po];
   if Pos('unit', Name) > 0 then
      PasEdit.Lines[Po]:= Format('%s %s;',['unit',GetNetFileName(UnitName)]);
end;

procedure TEditorForm.FormActivate(Sender: TObject);
begin
   FMainForm.ToggleFormUnit:= False;
end;

procedure TEditorForm.NewSheet(const FileName:string);
// create new editor
var
   PasEdit:TmwCustomEdit;
   TabSheet:TTabSheet;
   List:TStringList;
begin
   TabSheet:=TTabSheet.Create(PageControl1);
   with TabSheet do
   begin
      PageControl := PageControl1;
      if ExtractFileExt(FileName)='.pas' then Caption:= GetNetFileName(FileName)
      else Caption:= ExtractFileName(FileName);
      PasEdit            := TmwCustomEdit.Create(PageControl1);
      PasEdit.Parent     := TabSheet;
     // RichEdit.Name := UniqueName(RichEdit);
      PasEdit.Color      := IniFile.ReadInteger('EDITOR','BackColor',clWhite);
      PasEdit.Font.Name  := IniFile.ReadString('EDITOR','FONTNAME','Arial');
      PasEdit.Font.Size  := IniFile.ReadInteger('EDITOR','FONTSIZE',10);
      PasEdit.Align      := alClient;
      PasEdit.PopupMenu  := EditorLocalMenu;
      PasEdit.ScrollBars := ssBoth;
      PasEdit.HighLighter:= PasSyn;
      PasEdit.OnStatusChange := StatusChangeHandler;
   end;
   Caption:=FileName;
   PageControl1.ActivePage:=TabSheet;
   Show;
   ActivePasEditor:= PasEdit;
end;

procedure TEditorForm.ecClosePageClick(Sender: TObject);
begin
   FMainForm.CloseActiveFile;
end;

procedure TEditorForm.ecOpenFileAtCursorClick(Sender: TObject);
var
   List:TStringList;
   I:Integer;
begin
   if OpenDialog1.Execute then
   begin
      List:=TStringList.Create;
      List.LoadFromFile(OpenDialog1.FileName);
      for I:=0 to List.Count-1 do
          ActivePasEditor.Lines.Insert(1,List[I]);
   end;
end;

procedure TEditorForm.ecReadOnlyItemClick(Sender: TObject);
begin
   ecReadOnlyItem.Checked:=not ecReadOnlyItem.Checked;
   ActivePasEditor.ReadOnly:= ecReadOnlyItem.Checked;
   if ecReadOnlyItem.Checked then   StatusBar.Panels[1].Text:='ReadOnly'
   else StatusBar.Panels[1].Text:='';
end;

// Ŀ ġ ִ ܾ ˷ش.
function TEditorForm.GetCursorWord:string;
var
  WordText:string;
  I,LineLength,LinePos:Integer;
  CurPos,StartPos,EndPos:Integer;
begin
   Result:='';
   WordText:= ActivePasEditor.LineText;  // Ŀ ġ ִ  ڿ
   if Length(WordText)<1 then Exit;
   CurPos:=ActivePasEditor.CaretX;
   LineLength:= Length(WordText);
   for I:= CurPos downto 1 do
   begin
      if (WordText[I] ='')  then Exit;
      if WordText[I] in [' ',#0..#47,#58..#64,#91..#96,#123..#127] then Break;
   end;
   StartPos:= I+1;  // ܾ  ġ
   for I:= CurPos to LineLength do begin
      if WordText[I] ='' then Exit;
      if WordText[I] in [' ',#0..#47,#58..#64,#91..#96,#123..#127] then Break;
   end;
   EndPos:= I;  // ܾ  ġ
   WordText:= Trim(Copy(WordText,StartPos, EndPos-StartPos));
   Result:= WordText;
end;

procedure TEditorForm.CaretInfo;
var
   Col,Row:integer;
begin
   // ĳ ġ Ѵ.
   Col:= ActivePasEditor.CaretX;
   Row:= ActivePasEditor.CaretY;
   StatusBar.Panels[0].Text:=IntToStr(Row)+' :' + IntToStr(Col);
end;

procedure TEditorForm.Bookamrk01Click(Sender: TObject);
// Set BookMark
var
   BookMark, X, Y: integer;
   PasEdit: TmwCustomEdit;
begin
   (Sender as TMenuItem).Checked:= not (Sender as TMenuItem).Checked;
   if (Sender as TMenuItem).Checked then
   begin
      PasEdit:= ActivePasEditor;
      if not PasEdit.Gutter.Visible then PasEdit.Gutter.Visible:= True;
      BookMark:= (Sender as TMenuItem).Tag;
      X:= PasEdit.CaretX;
      Y:= PasEdit.CaretY;
      PasEdit.SetBookMark(BookMark, X, Y);
   end
   else ActivePasEditor.ClearBookMark((Sender as TMenuItem).Tag);
end;

procedure TEditorForm.Bookmark01Click(Sender: TObject);
begin
   ActivePasEditor.GotoBookMark((Sender as TMenuItem).Tag);
end;

procedure TEditorForm.ecTopicSearchClick(Sender: TObject);
begin
   ShowKeywordHelp(GetCursorWord);
end;

function  TEditorForm.GetLinkWithFile(const FileName:string): TmwCustomEdit;
var
  I: Integer;
  Name: string;
begin
   Result:= ActivePasEditor;
   for I:= Low(FMainForm.OpenUnitList) to High(FMainForm.OpenUnitList) do
   begin
      Name:= FMainForm.OpenUnitList[I].FileName;
      if CompareText(Name, FileName) = 0 then
      begin
         Result:= FMainForm.OpenUnitList[I].Editor;
         Break;
      end;
   end;
end;

function  TEditorForm.IsIncludeUnit(UnitName:string): Boolean;
var
  StartLine, EndLine, I: Integer;
begin
   Result:= False;
   ParserUnit.SetUnitSource(ActivePasEditor.Lines);
   StartLine:= ParserUnit.GetStrPos('uses');
   EndLine:= ParserUnit.GetStrPos('type');
   for I:= StartLine to EndLine do
      if Pos(UnitName, ParserUnit.UnitSource[I]) > 0 then
      begin
         Result:= True;
         Break;
      end;
end;

procedure TEditorForm.AddUseUnit(UnitName:string);
var
  StartLine, EndLine, I, Pos1: Integer;
begin
   if ParserUnit.UnitSource.Count = 0 then
      ParserUnit.SetUnitSource(ActivePasEditor.Lines);
   StartLine:= ParserUnit.GetStrPos('uses');
   EndLine:= ParserUnit.GetStrPos('type');
   for I:= StartLine to EndLine do
   begin
      Pos1:= Pos(';', ParserUnit.UnitSource[I]);
      if  Pos1 > 0 then
      begin
          ActivePasEditor.Lines[I]:= Copy(ActivePasEditor.Lines[I], 1, Pos1-1)
              +' ,'+ UnitName + ';';
          Break;
      end;
   end;
end;

procedure TEditorForm.CheckUseUnit(Component: TComponent);
// ҽ uses Component unitname ԵǾ   ԽŲ.
var
   Data: PTypeData;
   Info: PTypeInfo;
   UnitName: string;
begin
   Info:= Component.ClassInfo;
   Data:= GetTypeData(Info);
   UnitName:= Data.UnitName;
   if not IsIncludeUnit(UnitName) then AddUseUnit(UnitName);
end;

procedure TEditorForm.FormCreate(Sender: TObject);
begin
   ParserUnit:= TParserUnit.Create('');
   PasSyn.AsmAttri.Style       := [];
   PasSyn.CommentAttri.Style   := [];
   PasSyn.IdentifierAttri.Style:= [];
   PasSyn.KeyAttri.Style       := [];
   PasSyn.NumberAttri.Style    := [];
   PasSyn.SpaceAttri.Style     := [];
   PasSyn.StringAttri.Style    := [];
   PasSyn.SymbolAttri.Style    := [];
end;

procedure TEditorForm.FormDestroy(Sender: TObject);
begin
   ParserUnit.Free;
end;

procedure TEditorForm.ecViewAsFormItemClick(Sender: TObject);
var
   Input,Output:TMemoryStream;
   FileName: string;
   Form: TForm;
begin
   FileName:= FilePath + TempFile;
   ActivePasEditor.Lines.SaveToFile(FileName);
   Input:= TMemoryStream.Create;
   Input.LoadFromFile(FileName);
   Output:= TMemoryStream.Create;
   ObjectTextToResource(Input,Output);
   Output.SaveToFile(FilePath+TempForm);
   Output.LoadFromFile(FilePath+TempForm);

   Form:= CreateProxy(TProxyForm,GetNetFileName(FileName)) as TForm;
   Output.Position:=0;
   // ReadClass;
   try
      Output.ReadComponentRes(Form);
   finally
      Input.Free;
      OutPut.Free;
   end;
end;

procedure TEditorForm.MwCompletionProposal1Execute(Sender: TObject);
// code completion procedure
// find and display object's properties, methods, events
// see object.dat file
var
  Component: TComponent;
  Idx: Integer;
begin
   MwCompletionProposal1.ItemList.Clear;
   Component:= FMainForm.ActiveForm.FindComponent(MwCompletionProposal1.CurrentString);
   if Component <> nil then
   begin
      Idx:= CodeCompletionList.IndexOf('[' + Component.ClassName + ']');
      if Idx <> -1 then
      begin
         Inc(Idx);
         while (Idx < CodeCompletionList.Count) and
               (CodeCompletionList[Idx][1] <> '[') do
         begin
            MwCompletionProposal1.ItemList.Add(CodeCompletionList[Idx]);
            Inc(Idx);
         end;
      end;
   end
   else
   begin   // if not exists token, then active form's component list
      for Idx:= 0 to FMainForm.ActiveForm.ComponentCount - 1 do
      begin
         MwCompletionProposal1.ItemList.Add(
            Format('var     %s:%s', [FMainForm.ActiveForm.Components[Idx].Name,
                                     FMainForm.ActiveForm.Components[Idx].ClassName]));
      end;
   end;
end;

procedure TEditorForm.PasSynToken(Sender: TObject; TokenKind: Integer;
  TokenText: String; LineNo: Integer);
var
 _TokenKind                    : mwPasSyn.TtkTokenKind absolute TokenKind;
begin
   if not FMainForm.ViewCodeExplorerItem.Checked then Exit;
   if (_TokenKind = tkKey) and (Uppercase(TokenText) = 'CLASS') then
      ParseState := psInClassDef;

   if (_TokenKind = tkKey) and (Uppercase(TokenText) = 'END') and
      (ParseState = psInClassDef) then
      ParseState := psIdle;

   if (_TokenKind = tkKey) and (Uppercase(TokenText) = 'IMPLEMENTATION') then
      ObjectInspector.InsertIntoList(TokenText, LineNo, False);

   if ParseState <> psInClassDef then
   begin
      if (ParseState = psIdle) and  (_TokenKind = tkKey) and
         ((Uppercase(TokenText) = 'PROCEDURE') or (Uppercase(TokenText) = 'FUNCTION')) then
         ParseState := psProcSeen;

      if (ParseState = psProcSeen) and (_TokenKind = tkIdentifier) then
      begin
         ProcedureName := TokenText;
         ParseState := psWaitForSymbol;
      end;

      if (ParseState = psWaitForSymbol) and (_TokenKind = tkSymbol) and
         (TokenText = '.') then
         ParseState := psProcSeen;

      if (ParseState = psWaitForSymbol) and  (_TokenKind = tkSymbol) and
         (TokenText <> '.') then
      begin
         ObjectInspector.InsertIntoList(ProcedureName, LineNo, True);
         ParseState := psIdle;
      end;
   end;
end;

procedure TEditorForm.ecViewExplorerItemClick(Sender: TObject);
begin
   FMainForm.ViewCodeExplorerItemClick(self);
end;

procedure TEditorForm.CodeTemplateValidate(Sender: TObject);
var
  Code, CodeStr: string;
  Pos1: Integer;
begin
   Code:= CompletionProposalForm.ItemList[CompletionProposalForm.Position];
   Pos1:= Pos(' ', Code);
   if Pos1 > 0 then Code:= Copy(Code, 1, Pos1 - 1);
   CodeStr:= mwAutoComplete1.GetTokenValue(Code);
   if CodeStr='' then Exit;
   ActivePasEditor.SetSelStart(ActivePasEditor.GetSelStart - Length(Code));
   ActivePasEditor.SetSelEnd(ActivePasEditor.GetSelStart + Length(Code));
   ReplaceSelText(CodeStr);
   ActivePasEditor.SetFocus;
end;

procedure TEditorForm.ecViewCodeTemplateClick(Sender: TObject);
var
  p: TPoint;
begin
   CompletionProposalForm:= TCompletionProposalForm.Create(self);
   with CompletionProposalForm do
   begin
      ClSelect:= clAqua;
      ItemList.Text:= mwAutoComplete1.GetTokenList;
      with ActivePasEditor do
         p := ClientToScreen(Point(CaretXPix, CaretYPix+LineHeight));
      Left:= p.x;
      Top:= p.y;
      CompletionProposalForm.OnValidate:= CodeTemplateValidate;
      Show;
   end;
end;

procedure TEditorForm.ReplaceSelText(Replace:string);
begin
   ActivePasEditor.SelText:= Replace;
end;

procedure TEditorForm.MwCompletionProposal1CodeCompletion(
  var Value: String; var X: Integer);
var
  i, j, k: Integer;
begin
   i:= Pos(' ', Value);
   if i > 0 then
   begin
     j:= Pos(':', Value);   // property type position
     k:= Pos('(', Value);   // procedure parameters position
     if k > 0 then j:= k;
     if j > 0 then  Value:= Trim(Copy(Value, i + 1, j - i - 1))
     else Value:= Trim(Copy(Value, i + 1, Length(Value)));
   end;
end;

procedure TEditorForm.StatusChangeHandler(Sender: TObject;
  Changes: TmwStatusChanges);
const
  ModifiedStrs: array[boolean] of string = ('', 'Modified');
  InsertModeStrs: array[boolean] of string = ('Overwrite', 'Insert');
var
  Edit: TmwCustomEdit;
begin
  Edit:= GetLinkEditor;
  // mwscAll can be used to update everything, see OnCreate handler
  with StatusBar do begin
    // statuspanel 0: caret position
    if Changes * [mwscAll, mwscCaretX, mwscCaretY] <> [] then
    begin
      Panels[0].Text := Format('%6d:%3d', [Edit.CaretY, Edit.CaretX]);
    end;
    // statuspanel 1: Modified property
    if Changes * [mwscAll, mwscModified] <> [] then
      Panels[1].Text := ModifiedStrs[Edit.Modified];
    // statuspanel 2: InsertMode property
    if Changes * [mwscAll, mwscInsertMode] <> [] then
      Panels[2].Text := InsertModeStrs[Edit.InsertMode];
  end;
end;

end.


